// ConfigWriter.cpp: implementation of the CConfigWriter class.
//
//////////////////////////////////////////////////////////////////////

#include "pch.h"
#include "Common.h"
#include "Temp.hpp"
#include "Logger.h"
#include "ConfigWriter.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


CConfigWriter::CConfigWriter(const tstring& aFileName,
	LPCTSTR sourceEncoding /* = NULL */, LPCTSTR destEncoding /* = NULL */)
	: fileName(aFileName), closed(false), resolveEntities(false), resolvedData(),
	  convertBuffer(NULL), bufferLen(-1)
{
	// extract path
	tstring	path, prefix;

	int	lastSeparatorPos = fileName.find_last_of( PATH_SEPARATOR );
	if (lastSeparatorPos >= 0) {
		path   = fileName.substr( 0, lastSeparatorPos );
		prefix = fileName.substr( lastSeparatorPos + 1 );
	}
	else {
		// path is empty
		prefix = fileName;
	}

	tempFileName = getTempFileName( path, prefix );
	if (tempFileName.length() <= 0) {
		// TODO: raise exception
	}

	out.open( tempFileName.c_str() );
	// TODO: check result of file opening

	needConvertion = false;
	if (sourceEncoding != NULL && destEncoding != NULL) {
		if (lstrcmp( sourceEncoding, destEncoding ) != 0) {
			try {
				charSetConvertor.setTransformation( sourceEncoding, destEncoding );
				needConvertion = true;
			} catch (miniSAX::CMiniSAXException&) {
				// ensure that we will not use convertion
				needConvertion = false;
			}
		}
		// destination encoding specified => set it in document
		out << "<?xml version=\"1.0\" encoding=\"" << destEncoding << "\" ?>\n";
	}
}


CConfigWriter::~CConfigWriter()
{
	if (!closed) {
		close();
	}
	if (convertBuffer != NULL) {
		delete [] convertBuffer;
	}
}


void CConfigWriter::close() {
	if (!closed) {
		out.close();

		LoggerPtr	logger;
		logger << LOG_ERROR;

		if (_tunlink( fileName.c_str() ) == 0 || errno == ENOENT) {
			if (_trename( tempFileName.c_str(), fileName.c_str() ) == 0) {
				if (!::SetFileAttributes( fileName.c_str(), FILE_ATTRIBUTE_NORMAL )) {
					// we failed to set normal attribute
					logger << TEXT("failed to set normal attribute for the configuration file:\n")
						   << TEXT("  file name = \"") << fileName << TEXT("\"\n");
				}
			} else {
				// we failed to rename temp name to the real name
				logger << TEXT("failed to rename temporary file into real configuration file:\n")
					   << TEXT("  temp file name = \"") << tempFileName << TEXT("\"\n")
                       << TEXT("  real file name = \"") << fileName << TEXT("\"\n");
			}
		} else {
			// we failed to remove old configuration file
			logger << TEXT("failed to remove old configuration file:\n")
				   << TEXT("  file name = \"") << fileName << TEXT("\"\n");
		}

		closed = true;

	} else {
		// TODO: raise exception
	}
}


void CConfigWriter::cancel() {
	if (!closed) {
		out.close();
		remove( tempFileName.c_str() );
	}
}


CConfigWriter& CConfigWriter::operator <<(const tstring& data) {
	print( data.c_str() );

	return (*this);
}


CConfigWriter& CConfigWriter::operator <<(LPCTSTR data) {
	print( data );

	return (*this);
}


CConfigWriter& CConfigWriter::operator <<(const int data) {
	TCHAR	buffer[128];
	_stprintf( buffer, TEXT("%d"), data );
	print( buffer );

	return (*this);
}


CConfigWriter& CConfigWriter::operator <<(const CEntityResolver& resolver) {
	resolveEntities = true;
	return (*this);
}


void CConfigWriter::print(LPCTSTR data) {
	if (resolveEntities) {
		resolve( data );
		data = resolvedData.c_str();
		resolveEntities = false;
	}
	if (needConvertion) {
		data = encode( data );
	}

	out << data;
}


void CConfigWriter::resolve(LPCTSTR data) {
	// reset resolved data
	resolvedData = TEXT("");

	tstring	entity;
	int		startPos = 0;
	int		endPos   = 0;
	while (data[endPos] != TEXT('\0')) {
		bool	resolveTime = true;	// nice name? ;)
		switch (data[endPos]) {
		case '<':
			entity = TEXT("&lt;");
			break;
		case '>':
			entity = TEXT("&gt;");
			break;
		case '"':
			entity = TEXT("&quot;");
			break;
		case '\'':
			entity = TEXT("&apos;");
			break;
		case '&':
			entity = TEXT("&amp;");
			break;
		default:
			resolveTime = false;
		}
		if (resolveTime) {
			if (startPos < endPos) {
				resolvedData.append( data + startPos, endPos - startPos );
			}
			resolvedData += entity;
			++endPos;
			startPos = endPos;
		}
		else {
			++endPos;
		}
	}
	if (startPos < endPos) {
		resolvedData.append( data + startPos, endPos - startPos );
	}
}


char* CConfigWriter::encode(LPCTSTR data) {
	// grow buffer if it is not enough
	int	len = lstrlen( data );
	if (bufferLen < len) {
		if (convertBuffer != NULL) {
			delete [] convertBuffer;
		}
		bufferLen     = len;
		convertBuffer = new char[bufferLen+1];
	}
	charSetConvertor.convert( data, convertBuffer );
	return convertBuffer;
}
