#include "stdafx.h"

#include "StringHolder.h"
#include "CharBuffer.h"

#include "MiniSAXImpl.h"
#include "AttributeImpl.h"
#include "AttributeListImpl.h"

#define DEFINED(a) ((a) != NULL)


using namespace miniSAX;


/***********************************
  CMiniSAXParserImpl implementation
 ***********************************/

CMiniSAXParserImpl::CMiniSAXParserImpl() {
	// reset all handlers
	m_documentHandler = NULL;
	m_contentHandler  = NULL;
	m_errorHandler    = NULL;
	m_entityResolver  = NULL;

	m_currentRow    = -1;
	m_currentColumn = -1;

	// no convertor yet
	m_charSetConvertor = NULL;
}


CMiniSAXParserImpl::~CMiniSAXParserImpl()
{
	if (m_charSetConvertor != NULL) {
		delete m_charSetConvertor;
	}
}


bool CMiniSAXParserImpl::isGoodBeginning(CDataSource* dataSource) throw (CMiniSAXException) {
	if (!dataSource->hasData()) {
		handleError( CMiniSAXException( "unexpected end of data",
			dataSource->getColumn(), 
			dataSource->getRow() ) );
		return false;
	}
	if (dataSource->getChar() != '<') {
		handleError( CMiniSAXException( "unexpected symbol",
			dataSource->getColumn(), 
			dataSource->getRow() ) );
		return false;
	}

	return true;
}


void CMiniSAXParserImpl::parse(CDataSource* dataSource) throw(CMiniSAXException)
{
	// prepare buffer
	CCharBuffer	buffer( 1024 );
	CCharBuffer	storage( 1024 );	// will be used to resolve entities

	// start document parsing
	startDocument();

	// first symbol should be '<'
	if (!isGoodBeginning( dataSource )) {
		return;
	}

	// remember position
	m_currentRow    = dataSource->getRow();
	m_currentColumn = dataSource->getColumn();

	// get position into local variables
	int	col = dataSource->getColumn();
	int	row = dataSource->getRow();

	// analize next char, it can be '?" ("<?xml" for xml version and encoding)

	if (dataSource->getChar() == '?') {
		if (!processXMLDeclaration( dataSource )) {
			return;
		}
		dataSource->skipSpaces();
		if (!isGoodBeginning( dataSource )) {
			return;
		}
	} else {
		dataSource->back();
	}	

	// check comment at the beginning of file
	if (dataSource->getChar() == '!') {
		if (!skipComment( dataSource )) {
			return;
		}
		dataSource->skipSpaces();
		if (!isGoodBeginning( dataSource )) {
			return;
		}
	} else {
		dataSource->back();
	}

	if (!readStartOfElement(dataSource)) {
		return;
	}

	bool	noError = true;
	do {
		// remember position
		m_currentRow    = dataSource->getRow();
		m_currentColumn = dataSource->getColumn();

		int	distance = dataSource->getDistance( '<' );
		if (distance > 0) {
			// increase buffers if have to (+1 for '\0')
			buffer.checkSize( distance + 1 );
			storage.checkSize( distance + 1 );

			dataSource->getSequence( buffer, distance );
			buffer(distance) = '\0';

			resolveEntities( buffer, storage );

			if (m_charSetConvertor != NULL) {
				m_charSetConvertor->convert( buffer, buffer );
			}

			// handle characters
			handleCharacters( buffer, strlen( buffer ) );
		}
		
		// remember position again
		m_currentRow    = dataSource->getRow();
		m_currentColumn = dataSource->getColumn();

		if (dataSource->hasData()) {

			// we are definitly at '<'

			// determine "</" (end of element tag) or "<!--" (start of comment)
			bool	endOfElement = false;
			bool	comment      = false;
			// read '<'
			dataSource->getChar();
			if (dataSource->hasData()) {
				switch (dataSource->getChar()) {
					case '/':
						endOfElement = true;
						break;
					case '!':
						comment = true;
						break;
					default:
						dataSource->back();
						break;
				}
			}
			if (comment) {
				noError = skipComment( dataSource );
			} else {
				if (endOfElement) {
					noError = readEndOfElement(dataSource);
				}
				else {
					noError = readStartOfElement(dataSource);
				}
			}
		}
	} while (noError && dataSource->hasData());

	// end document parsing
	endDocument();
}


bool CMiniSAXParserImpl::processXMLDeclaration(CDataSource* dataSource) {
	// on entering this method we should be at "xml ... ?>"
	char*	xml;
	if (!readName( dataSource, xml )) {
		return false;
	}

	// holder will release memory
	CStringHolder	stringHolder(xml);

	if (strcmp( xml, "xml" ) != 0) {
		return false;
	}

	// read attributes

	CAttributeListImpl	attributes;

	bool	haveAllData  = false;
	do {
		dataSource->skipSpaces();
		// check close bracket of element
		if (!dataSource->hasData()) {
			handleError( CMiniSAXException( "unexpected end of data", 
											dataSource->getColumn(),
											dataSource->getRow() ) );
			return false;
		}
		switch (dataSource->getChar()) {
		case '?':
			{
				int	col = dataSource->getColumn();
				int	row = dataSource->getRow();
				if (!dataSource->hasData()) {
					handleError( CMiniSAXException( "unexpected end of xml declaration", col, row ) );
					return false;
				}
				if (dataSource->getChar() != '>') {
					handleError( CMiniSAXException( "unexpected symbol", col, row ) );
					return false;
				}
				haveAllData = true;
			}
			break;
		default:
			dataSource->back();
		}

		if (!haveAllData) {
			// read attribute name
			char*	attrName;
			if (!readName( dataSource, attrName )) {
				return false;
			}

			// holder will release memory
			CStringHolder	attrNameHolder( attrName );

			dataSource->skipSpaces();

			// check '='
			int	col = dataSource->getColumn();
			int	row = dataSource->getRow();
			if (!dataSource->hasData()) {
				handleError( CMiniSAXException( "unexpected end of data", col, row ) );
				return false;
			}
			if (dataSource->getChar() != '=') {
				handleError( CMiniSAXException( "unexpected symbol", col, row ) );
				return false;
			}

			dataSource->skipSpaces();

			// read attribute value
			char*	attrValue;
			if (!readValue( dataSource, attrValue )) {
				return false;
			}

			// holder will release memory
			CStringHolder	attrValueHolder( attrValue );

			// create attribute
			CAttributeImpl	attr( attrName, attrValue );

			attributes.addAttribute( attr );
		}
	} while (!haveAllData);	

	const CAttribute*	attr = attributes.getAttribute( "encoding" );
	if (attr != NULL) {
		m_documentEncoding = attr->getValue();
		if (m_handlerEncoding.length() > 0) {
			if (strcmp( m_documentEncoding.c_str(), m_handlerEncoding.c_str() ) != 0) {
				m_charSetConvertor = new CCharSetConvertor();
				try {
					m_charSetConvertor->setTransformation( m_documentEncoding.c_str(), m_handlerEncoding.c_str() );
				} catch (CMiniSAXException& /* ex */) {
					// handleError( ex );
					// return false;
					
					// we doesn't handle this error, we simply ignore it
					// TODO: notify user that we cannot make transformation

					delete m_charSetConvertor;
					m_charSetConvertor = NULL;
				}
			}
		}
	}

	return true;
}


bool CMiniSAXParserImpl::skipComment(CDataSource* dataSource) {
	// when this method is called dataSource should be after "<!"
	// it means that we skip all including "-->"
	// and we should check if we are on starting "--"
	if (!dataSource->hasData() || dataSource->getChar() != '-' || !dataSource->hasData()
		|| dataSource->getChar() != '-')
	{
		handleError( CMiniSAXException( "Comment start expected (<!--)", m_currentColumn, m_currentRow ) );
		return false;
	}
	// we'll use trigger with initial value of 0 and resulting value of 3
	// it increases each time good symbol found from "-->" and decreases on bad symbols
	// not quick but should work :)
	int	trigger = 0;
	while (trigger < 3 && dataSource->hasData()) {
		char	symbol = dataSource->getChar();
		switch (trigger) {
			case 0:
				if (symbol == '-') {
					++trigger;
				}
				break;
			case 1:
				if (symbol == '-') {
					++trigger;
				} else {
					--trigger;
				}
				break;
			case 2:
				if (symbol == '>') {
					++trigger;
				} else {
					--trigger;
				}
				break;
		}
	}
	if (trigger != 3) {
		handleError( CMiniSAXException( "End of comment not found (-->)", m_currentColumn, m_currentRow ) );
		return false;
	}

	return true;
}


/*

void CMiniSAXParser::parse(CDataSource* dataSource) throw(CMiniSAXException)
{
	// prepare buffer
	CCharBuffer	buffer( 1024 );
	CCharBuffer	storage( 1024 );	// will be used to resolve entities

	try {
		// start document parsing
		startDocument();

		// first symbol should be '<'

		// remember position
		currentRow    = dataSource->getRow();
		currentColumn = dataSource->getColumn();

		// get position into local variables
		int	col = dataSource->getColumn();
		int	row = dataSource->getRow();

		if (!dataSource->hasData()) {
			handleError( CMiniSAXException( "unexpected end of data", col, row ) );
			return;
		}

		if (dataSource->getChar() != '<') {
			handleError( CMiniSAXException( "unexpected symbol", col, row ) );
			return;
		}

		if (!readStartOfElement(dataSource)) {
			return;
		}

		bool	noError = true;
		do {
			// remember position
			currentRow    = dataSource->getRow();
			currentColumn = dataSource->getColumn();

			int	distance = dataSource->getDistance( '<' );
			if (distance > 0) {
				// increase buffers if have to (+1 for '\0')
				buffer.checkSize( distance + 1 );
				storage.checkSize( distance + 1 );

				dataSource->getSequence( buffer, distance );
				buffer(distance) = '\0';

				resolveEntities( buffer, storage );

				// handle characters
				handleCharacters( buffer, strlen( buffer ) );
			}
			
			// remember position again
			currentRow    = dataSource->getRow();
			currentColumn = dataSource->getColumn();

			if (dataSource->hasData()) {
				// determine '<''/'
				bool	endOfElement = false;
				// read '<'
				dataSource->getChar();
				if (dataSource->hasData()) {
					if (dataSource->getChar() == '/') {
						endOfElement = true;
					}
					else {
						dataSource->back();
					}
				}
				if (endOfElement) {
					noError = readEndOfElement(dataSource);
				}
				else {
					noError = readStartOfElement(dataSource);
				}
			}
		} while (noError && dataSource->hasData());

		// end document parsing
		endDocument();

	} catch(...) {
		buffer.reset();

		throw;
	}
}

*/

bool CMiniSAXParserImpl::readStartOfElement(CDataSource* dataSource) {
	dataSource->skipSpaces();

	// read name

	char*	name;
	if (!readName( dataSource, name )) {
		return false;
	}

	// holder will release memory
	CStringHolder	nameHolder(name);

	// read attributes

	CAttributeListImpl	attributes;

	bool	haveAllData  = false;
	bool	endOfElement = false;
	do {
		dataSource->skipSpaces();
		// check close bracket of element
		if (!dataSource->hasData()) {
			handleError( CMiniSAXException( "unexpected end of data", 
											dataSource->getColumn(),
											dataSource->getRow() ) );
			return false;
		}
		switch (dataSource->getChar()) {
		case '>':
			haveAllData = true;
			break;
		case '/':
			{
				int	col = dataSource->getColumn();
				int	row = dataSource->getRow();
				if (!dataSource->hasData()) {
					handleError( CMiniSAXException( "unexpected end of data", col, row ) );
					return false;
				}
				if (dataSource->getChar() != '>') {
					handleError( CMiniSAXException( "unexpected symbol", col, row ) );
					return false;
				}
				// it's the end of the element
				endOfElement = true;
				haveAllData = true;
			}
			break;
		default:
			dataSource->back();
		}

		if (!haveAllData) {
			// read attribute name
			char*	attrName;
			if (!readName( dataSource, attrName )) {
				return false;
			}

			// holder will release memory
			CStringHolder	attrNameHolder( attrName );

			dataSource->skipSpaces();

			// check '='
			int	col = dataSource->getColumn();
			int	row = dataSource->getRow();
			if (!dataSource->hasData()) {
				handleError( CMiniSAXException( "unexpected end of data", col, row ) );
				return false;
			}
			if (dataSource->getChar() != '=') {
				handleError( CMiniSAXException( "unexpected symbol", col, row ) );
				return false;
			}

			dataSource->skipSpaces();

			// read attribute value
			char*	attrValue;
			if (!readValue( dataSource, attrValue )) {
				return false;
			}

			// holder will release memory
			CStringHolder	attrValueHolder( attrValue );

			// create attribute
			CAttributeImpl	attr( attrName, attrValue );

			attributes.addAttribute( attr );
		}
	} while (!haveAllData);

	startElement( name, attributes );
	if (endOfElement) {
		endElement( name );
	}

	return true;
}


bool CMiniSAXParserImpl::readEndOfElement(CDataSource* dataSource) {
	dataSource->skipSpaces();

	// read name

	char*	name;
	if (!readName( dataSource, name )) {
		return false;
	}

	// holder will release memory on exit
	CStringHolder	nameHolder(name);

	dataSource->skipSpaces();
	int	col = dataSource->getColumn();
	int	row = dataSource->getRow();
	if (!dataSource->hasData()) {
		handleError( CMiniSAXException( "unexpected end of data", col, row ) );
		return false;
	}
	if (dataSource->getChar() != '>') {
		handleError( CMiniSAXException( "unexpected symbol", col, row ) );
		return false;
	}

	endElement( name );

	return true;
}


bool CMiniSAXParserImpl::readName(CDataSource* dataSource, char*& name) {
	int		col       = dataSource->getColumn();
	int		row       = dataSource->getRow();
	char	firstChar = dataSource->getChar();
	if (!isalpha(firstChar) && firstChar != '_') {
		handleError( CMiniSAXException( "unexpected symbol", col, row ) );
		return false;
	}

	size_t	distance = dataSource->getNonElementDistance();

	char*	buffer = new char[distance+2];	// + first char + '\0'
	if (distance > 0) {
		dataSource->getSequence( buffer + 1, distance );
	}
	buffer[0]          = firstChar;
	buffer[distance+1] = '\0';

	// we have name
	name = buffer;

	if (m_charSetConvertor != NULL) {
		m_charSetConvertor->convert( name, name );
	}

	return true;
}


bool CMiniSAXParserImpl::readValue(CDataSource* dataSource, char*& value) {
	int		col       = dataSource->getColumn();
	int		row       = dataSource->getRow();
	char	firstChar = dataSource->getChar();
	if (firstChar != '\'' && firstChar != '\"') {
		handleError( CMiniSAXException( "unexpected symbol", col, row ) );
		return false;
	}

	size_t	distance = dataSource->getDistance( firstChar );

	char*	buffer = new char[distance+1];	// + '\0'
	if (distance > 0) {
		dataSource->getSequence( buffer, distance );
	}
	buffer[distance] = '\0';

	char*	storage = new char[distance+1];
	resolveEntities( buffer, storage );
	delete [] storage;

	if (!dataSource->hasData()) {
		handleError( CMiniSAXException( "unexpected end of data", 
									    dataSource->getColumn(),
										dataSource->getRow() ) );
		return false;
	}

	// read closing quote
	dataSource->getChar();

	// we have value
	value = buffer;

	if (m_charSetConvertor != NULL) {
		m_charSetConvertor->convert( value, value );
	}

	return true;
}


void CMiniSAXParserImpl::startDocument() {
	if (DEFINED(m_documentHandler)) {
		m_documentHandler->startDocument();
	}
}


void CMiniSAXParserImpl::endDocument() {
	if (DEFINED(m_documentHandler)) {
		m_documentHandler->endDocument();
	}
}


void CMiniSAXParserImpl::startElement(const char* name, CAttributeList& attributes) {
	if (DEFINED(m_contentHandler)) {
		m_contentHandler->startElement( name, attributes );
	}
}


void CMiniSAXParserImpl::endElement(const char* name) {
	if (DEFINED(m_contentHandler)) {
		m_contentHandler->endElement( name );
	}
}


void CMiniSAXParserImpl::handleCharacters(const char* buffer, size_t len) {
	if (DEFINED(m_contentHandler)) {
		m_contentHandler->handleCharacters( buffer, len );
	}
}


void CMiniSAXParserImpl::handleError(const CMiniSAXException& e) {
	if (DEFINED(m_errorHandler)) {
		m_errorHandler->handleError( e );
	} else {
		throw CMiniSAXException(e);
	}
}


void CMiniSAXParserImpl::resolveEntities(char* data, char* storage) {
	int	pos = 0;
	while (data[pos] != '\0') {
		if (data[pos] == '&') {
			++pos;
			if (strncmp( data + pos, "lt;", 3 ) == 0) {
				strcpy( storage, data + pos + 3 );
				data[pos-1] = '<';
				strcpy( data+pos, storage );
				continue;
			}
			if (strncmp( data + pos, "gt;", 3 ) == 0) {
				strcpy( storage, data + pos + 3 );
				data[pos-1] = '>';
				strcpy( data+pos, storage );
				continue;
			}
			if (strncmp( data + pos, "apos;", 5 ) == 0) {
				strcpy( storage, data + pos + 5 );
				data[pos-1] = '\'';
				strcpy( data+pos, storage );
				continue;
			}
			if (strncmp( data + pos, "quot;", 5 ) == 0) {
				strcpy( storage, data + pos + 5 );
				data[pos-1] = '"';
				strcpy( data+pos, storage );
				continue;
			}
			if (strncmp( data + pos, "amp;", 4 ) == 0) {
				strcpy( storage, data + pos + 4 );
				data[pos-1] = '&';
				strcpy( data+pos, storage );
				continue;
			}
			// TODO: we've met '&' but failed to determine entity, we should raise an exception
		} else {
			++pos;
		}
	}
}


void CMiniSAXParserImpl::raiseEx(const char* msg) const throw(CMiniSAXException) {
	throw CMiniSAXException( msg, m_currentColumn, m_currentRow );
}


void CMiniSAXParserImpl::setHandlerEncoding(const char* encoding) throw(CMiniSAXException) {
	m_handlerEncoding = encoding;
}


const char* CMiniSAXParserImpl::getDocumentEncoding() const {
	if (m_documentEncoding.length() > 0) {
		return m_documentEncoding.c_str();
	} else {
		return NULL;
	}
}

void CMiniSAXParserImpl::checkUnexpected(CAttributeList& list) const throw(CMiniSAXException) {
	try {
		list.checkUnexpected();
	} catch (CMiniSAXException& e) {
		raiseEx( e.getMessage() );
	}
}
